/*----------------------------------------------------------------------------

  Susie Plugin for JPEG2000

  Copyright(C) 2008 fok all rights reserved.

----------------------------------------------------------------------------*/
#define STRICT
#include <windows.h>
#include <boost/cstdint.hpp>
using namespace boost;
#include <jasper/jasper.h>

/*----------------------------------------------------------------------------
	XLCTCYvZ}N
----------------------------------------------------------------------------*/
#define WIDTHBYTES(i) (((i)+31)/32*4)

/*----------------------------------------------------------------------------
	SPIPICTUREINFO
----------------------------------------------------------------------------*/
struct SPIPICTUREINFO
{
	LONG nLeft;				// 摜WJʒu
	LONG nTop;				// 摜WJʒu
	LONG nImageWidth; 			// 摜̕(pixel)
	LONG nImageHeight;			// 摜̍(pixel)
	WORD wXDensity; 		// f̐x
	WORD wYDensity; 		// f̐x
	SHORT nBitCount;		// fbit
	HLOCAL hTextInfo;		// 摜̃eLXg
};

/*----------------------------------------------------------------------------
	G[R[h
----------------------------------------------------------------------------*/
#define SPIERROR_NONE				0	// I
#define SPIERROR_NOTIMPLEMENTED		-1	// ̋@\̓CvgĂȂ
#define SPIERROR_USERCANCEL			1	// R[obN֐0Ԃ̂œWJ𒆎~
#define SPIERROR_FORMAT				2	// m̃tH[}bg
#define SPIERROR_DATA				3	// f[^Ă
#define SPIERROR_MEMORYALLOC		4	// mۂłȂ
#define SPIERROR_MEMORYACCESS		5	// G[(LockoȂA)
#define SPIERROR_FILEACCESS			6	// t@CG[
#define SPIERROR_UNKNOWN			7	// \isj

/*----------------------------------------------------------------------------
	R[obN
----------------------------------------------------------------------------*/
typedef BOOL (CALLBACK *SPIPROGRESSPROC)(INT, INT, LPARAM);

/*----------------------------------------------------------------------------
	JPEG2000f[^`FbN
----------------------------------------------------------------------------*/
bool CheckJpeg2000Format(LPVOID lpHeader, DWORD dwHeaderSize)
{
	jas_stream_t* stream = NULL;
	INT nDataFormat1;
	INT nDataFormat2;

	// libjasper
	jas_init();

	// Xg[J
	stream = jas_stream_memopen((char*)lpHeader, dwHeaderSize);
	if(stream == NULL)
	{
		jas_image_clearfmts();
		jas_cleanup();
		return false;
	}

	// C[W`𔻒
	nDataFormat1 = jas_image_strtofmt("jp2");
	nDataFormat2 = jas_image_getfmt(stream);

	jas_stream_close(stream);
	jas_image_clearfmts();
	jas_cleanup();
	return (nDataFormat2 == nDataFormat1);
}

/*----------------------------------------------------------------------------
	JPEG2000摜擾
----------------------------------------------------------------------------*/
bool GetJpeg2000Info(LPCSTR szFileName, SPIPICTUREINFO* lpInfo)
{
	jas_stream_t* stream = NULL;
	INT nDataFormat1;
	INT nDataFormat2;

	// libjasper
	jas_init();

	// Xg[J
	stream = jas_stream_fopen(szFileName, "rb");
	if(stream == NULL)
	{
		jas_image_clearfmts();
		jas_cleanup();
		return false;
	}

	// C[W`𔻒
	nDataFormat1 = jas_image_strtofmt("jp2");
	nDataFormat2 = jas_image_getfmt(stream);
	if(nDataFormat2 != nDataFormat1)
	{
		return false;
	}

	bool bRet = true;
	if(lpInfo)
	{
		// 摜𓾂
		jas_image_t* img = jas_image_decode(stream, nDataFormat2, NULL);
		if(!img)
		{
			bRet = false;
		}
		else
		{
			INT nImageWidth = jas_image_cmptwidth(img, 0);
			INT nImageHeight = jas_image_cmptheight(img, 0);
			INT nImageDepth = jas_image_cmptprec(img, 0);
			if(nImageWidth < 1 || nImageHeight < 1 ||
				nImageWidth > 32767 || nImageHeight > 32767 || nImageDepth != 8)
			{
				bRet = false;
			}
			else
			{
				// ǂݎ萬
				lpInfo->nImageWidth = nImageWidth;
				lpInfo->nImageHeight = nImageHeight;
				lpInfo->nBitCount = 24;
			}
			jas_image_destroy(img);
		}
	}

	jas_stream_close(stream);
	jas_image_clearfmts();
	jas_cleanup();
	return bRet;
}

/*----------------------------------------------------------------------------
	JPEG2000f[^fR[h
----------------------------------------------------------------------------*/
bool ReadJpeg2000(LPCSTR szFileName, HLOCAL* phHeader, HLOCAL* phImage)
{
	jas_stream_t* stream = NULL;
	jas_image_t* img = NULL;
	jas_matrix_t* rows[4] = {NULL,NULL,NULL,NULL};
	INT nDataFormat1;
	INT nDataFormat2;
	INT nChannels;
	INT nBits;
	INT nWidthBytes;
	INT nPixelDepth;
	INT nImageWidth, nImageHeight;
	INT x, y, nChannelIndex;
	LPBYTE lpDIBLine, lpDIBBits;
	BITMAPINFOHEADER* lpbi;
	bool bRet = false;

	// libjasper
	jas_init();

	// t@Cǂݍ
	stream = jas_stream_fopen(szFileName, "rb");
	if(stream == NULL)
	{
		jas_image_clearfmts();
		jas_cleanup();
		return false;
	}

	// C[W`
	nDataFormat1 = jas_image_strtofmt("jp2");
	nDataFormat2 = jas_image_getfmt(stream);
	if(nDataFormat2 != nDataFormat1)
	{
		jas_stream_close(stream);
		jas_image_clearfmts();
		jas_cleanup();
		return false;
	}

	// 摜fR[h
	try
	{
		img = jas_image_decode(stream, nDataFormat2, NULL);
	}
	catch(...)
	{
		img = NULL;
	}
	if(!img)
	{
		jas_stream_close(stream);
		jas_image_clearfmts();
		jas_cleanup();
		return false;
	}

	// ǂݍ݊
	jas_stream_close(stream);

	// `l擾
	nChannels = jas_image_numcmpts(img);

	// 1E3E4`l̂ݑΉ
	switch(nChannels)
	{
	case 1:
	case 3:
	case 4:

		nBits = 24;
		break;

	default:

		goto RaisedException;
	}

	// 摜𓾂
	nImageWidth = jas_image_cmptwidth(img, 0);
	nImageHeight = jas_image_cmptheight(img, 0);
	nPixelDepth = jas_image_cmptprec(img, 0);
	nWidthBytes = WIDTHBYTES(nImageWidth * 24);
	if(nImageWidth < 1 || nImageHeight < 1 ||
		nImageWidth > 32767 || nImageHeight > 32767 ||
		nPixelDepth != 8)
	{
		goto RaisedException;
	}

	// C[W
	*phHeader = LocalAlloc(LPTR, sizeof(BITMAPINFOHEADER));
	if(!(*phHeader))
	{
		goto RaisedException;
	}
	*phImage = LocalAlloc(LPTR, nWidthBytes * nImageHeight);
	if(!(*phImage))
	{
		LocalFree(*phHeader);
		goto RaisedException;
	}

	// DIBwb_
	lpbi = reinterpret_cast<BITMAPINFOHEADER*>(*phHeader);
	lpDIBBits = reinterpret_cast<LPBYTE>(*phImage);
    lpbi->biSize = sizeof(BITMAPINFOHEADER);
    lpbi->biWidth = nImageWidth;
    lpbi->biHeight = nImageHeight;
    lpbi->biPlanes = 1;
    lpbi->biBitCount = 24;
    lpbi->biCompression = BI_RGB;

	// f[^
	for(nChannelIndex = 0; nChannelIndex < nChannels; nChannelIndex++)
	{
		if ((rows[nChannelIndex] = jas_matrix_create(1, nImageWidth)) == NULL)
		{
			goto RaisedException;
		}
	}

	// C[WDIBɕϊJn
	for(y=0; y<nImageHeight; y++)
	{
		// e`lǂݍ
		for(nChannelIndex = 0; nChannelIndex < nChannels; nChannelIndex++)
		{
			if(jas_image_readcmpt(img, nChannelIndex, 0, y, nImageWidth, 1, rows[nChannelIndex]))
			{
				goto RaisedException;
			}
		}

		// DIBf[^ƂĊi[
		lpDIBLine = lpDIBBits + nWidthBytes * (nImageHeight - 1 - y);
		for (x=0; x<nImageWidth; x++)
		{
			if(nChannelIndex == 1)
			{
				// O[XP[
				lpDIBLine[x * 3    ] =
				lpDIBLine[x * 3 + 1] =
				lpDIBLine[x * 3 + 2] = BYTE(jas_matrix_getv(rows[0], x));
			}
			else
			{
				// RGBf[^
				lpDIBLine[x * 3    ] = BYTE(jas_matrix_getv(rows[2], x));
				lpDIBLine[x * 3 + 1] = BYTE(jas_matrix_getv(rows[1], x));
				lpDIBLine[x * 3 + 2] = BYTE(jas_matrix_getv(rows[0], x));
			}
		}
	}

	bRet = true;

RaisedException:

	// 
	for(nChannelIndex = 0; nChannelIndex < nChannels; nChannelIndex++)
	{
		if(rows[nChannelIndex] != NULL)
		{
			jas_matrix_destroy(rows[nChannelIndex]);
		}
	}

	jas_image_destroy(img);
	jas_image_clearfmts();
	jas_cleanup();
	return bRet;
}

/*----------------------------------------------------------------------------
	APIFvOC擾
----------------------------------------------------------------------------*/
EXTERN_C INT WINAPI GetPluginInfo(INT nIndex, LPSTR szRetBuffer, INT nBufLength)
{
	INT nSize;
	LPCSTR apszPluginInfo[] =
	{
		/* 0 */     "00IN",
		/* 1 */     "JPEG2000 loader Ver 1.00",
		/* 2n+1 */  "*.jp2;*.j2k",
		/* 2n+2 */  "JPEG2000 (*.jp2;*.j2k)",
	};

	if(nIndex < _countof(apszPluginInfo))
	{
		LPCSTR s = apszPluginInfo[nIndex];
		nSize = lstrlen(s) + 1;
		nSize = (nSize < nBufLength) ? nSize : nBufLength;
		CopyMemory(szRetBuffer, apszPluginInfo[nIndex], nSize);
	}
	else
	{
		nSize = 0;
	}

	return nSize;
}

/*----------------------------------------------------------------------------
	APIFt@CT|[g`FbN
----------------------------------------------------------------------------*/
EXTERN_C BOOL WINAPI IsSupported(LPCSTR szFileName, DWORD dwFlag)
{
	BYTE abyHeader[2000];

	/*
	dwFlag:
	ʃ[h0		t@Cnh
	ʃ[h0	t@C擪(2Kbyteȏ)ǂݍ񂾃obt@ւ̃|C^B
						t@CTCY2Kbyteȉ̏ꍇobt@2KbytemۂA
						]0Ŗ߂邱ƁB
	*/

	if(HIWORD(dwFlag) == 0)
	{
		// dwFlag : t@Cnh
		DWORD dwRead = 0;
		ZeroMemory(abyHeader, sizeof(abyHeader));
		ReadFile((HANDLE)dwFlag, abyHeader, sizeof(abyHeader), &dwRead, NULL);
	}
	else
	{
		// dwFlag : obt@
		CopyMemory(abyHeader, (LPVOID)dwFlag, sizeof(abyHeader));
	}

	if(!CheckJpeg2000Format(abyHeader, sizeof(abyHeader)))
	{
		return FALSE;
	}
	return TRUE;
}

/*----------------------------------------------------------------------------
	APIFt@C摜擾
----------------------------------------------------------------------------*/
EXTERN_C INT WINAPI GetPictureInfo(LPVOID lpFileData, LONG lLength, UINT uFlag, SPIPICTUREINFO* lpInfo)
{
	return SPIERROR_NOTIMPLEMENTED;
}

/*----------------------------------------------------------------------------
	APIFt@C摜ǂݍ
----------------------------------------------------------------------------*/
EXTERN_C INT WINAPI GetPicture(LPVOID lpFileData, LONG lLength, UINT uFlag, HLOCAL* phHeader, HLOCAL* phImage, SPIPROGRESSPROC lpCallback, LPARAM lParam)
{
	// nꂽf[^̏
	switch(uFlag & 0x07)
	{
	default:

		return SPIERROR_NOTIMPLEMENTED;

	case 0: // file

		// lpFileData	: LPCSTR t@C
		// lLength		: ǂݍ݊JnItZbg
		break;

	case 1: // memory

		return SPIERROR_NOTIMPLEMENTED;
	}

	// WJJn
	if(!ReadJpeg2000((LPCSTR)lpFileData, phHeader, phImage))
	{
		return SPIERROR_DATA;
	}

	return SPIERROR_NONE;
}

/*----------------------------------------------------------------------------
	APIFt@CTlC
----------------------------------------------------------------------------*/
EXTERN_C INT WINAPI GetPreview(LPVOID lpFileData, LONG lLength, UINT uFlag, HLOCAL* phHeader, HLOCAL* phImage, SPIPROGRESSPROC lpCallback, LPARAM lParam)
{
	return SPIERROR_NOTIMPLEMENTED;
}

/*----------------------------------------------------------------------------
	APIFo[WIvV_CAO\
----------------------------------------------------------------------------*/
EXTERN_C INT WINAPI ConfigurationDlg(HWND hWnd, INT nIndex)
{
	/*
	Susiěݒɂă_CAO\[X؂ւɂ
	ȉ̃WXg̒l擾Ĕf(Win32sł̓WXgł͂Ȃ
	INIt@Cgp)B
	WXgf[^݂Ȃꍇu1v̏ꍇɂ́u{vA
	u0v̏ꍇɂ́upvƂďB

	Win95/NT(Win32)	: HKEY_CURRENT_USER\Software\Takechin\Susie\Resource / REG_DWORD / Japanese
	3.1+Win32s		: SUSIE.INI / section [Resource] / key [Japanese]
	*/

	switch(nIndex)
	{
	case 0: // About

		//MessageBox(hWnd, "SPI Sample", "About", MB_OK);
		return SPIERROR_NOTIMPLEMENTED;

	case 1: // Option

		MessageBox(hWnd, "ݒ͂܂B", "Susie Plugin for JPEG2000", MB_OK);
		return SPIERROR_NONE;
	}

	return SPIERROR_NOTIMPLEMENTED;
}
